#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <libgen.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <net/if.h>
#include <getopt.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>

#include <linux/can.h>
#include <linux/can/raw.h>
#include <linux/can/error.h>

/* Application infos */
#define APP_NAME            "can_test"
#define APP_VERSION            "2.1"

#define pr_error(...)            do { myprintf(ERROR, __VA_ARGS__); } while(0)
#define pr_warning(...)            do { myprintf(WARNING, __VA_ARGS__); } while(0)
#define pr_info(...)            do { myprintf(INFO, __VA_ARGS__); } while(0)
#define pr_debug(...)            do { myprintf(DEBUG, __VA_ARGS__); } while(0)
#define pr_naked(...)            do { printf(__VA_ARGS__); } while(0)

#define CAN_FRAME_SIZE            sizeof(struct can_frame)

/* Operation modes */
typedef enum mode {
    MODE_RECEIVER = 0,
    MODE_TRANSMITTER,
} mode_t1;

/* Used for the configuration varaible st.verbose */
typedef enum verbosity_t {
    ERROR = 0,
    WARNING,
    INFO,
    DEBUG,
} verbosity_t;
static verbosity_t selected_verbosity;

/* Internal structure */
struct opts_t {
    char *iface;
    canid_t *canids;
    double *tx_rates;
    int ids;
    int extended;
    unsigned char pattern;
    mode_t1 mode;
    int bytes;
    int timeout;
    int secure;
    unsigned long loops;
    int xdelay;
};

static struct opts_t *main_opts;

inline static double now(void) {
    struct timeval tp;
    gettimeofday(&tp, NULL);
    return ((double)tp.tv_sec + (double)tp.tv_usec * 1e-6);
}

inline static void myprintf(verbosity_t level, const char *format, ...) {
    va_list lst;
    char *marke;
    int weg = 0;
    if (level > selected_verbosity) {
        return;
    }
    switch (level) {
    case ERROR:
        marke = "ERROR";
        weg = 1;
        break;
    case WARNING:
        marke = "WARNING";
        break;
    case INFO:
        marke = "INFO";
        break;
    case DEBUG:
        marke = "DEBUG";
        break;
    default:
        marke = "UNKNOW";
        weg = 1;
        break;
    }
    printf("[ %s ] ", marke);
    va_start(lst, format);
    vprintf(format, lst);
    va_end(lst);
    if (weg)
        fflush(stdout);
}

static void print_version(void) {
    pr_naked("CAN test application v%s\n", APP_VERSION);
}

static void print_usage(void) {
    fprintf(stdout, "Usage: %s [OPTIONS]\n"
        "%s %s Copyright Digi International Inc.\n\n"
        "Data transfer using CAN-sockets\n"
        "\n"
        "  -m, --master       Run the test as master (trasmitter)\n"
        "  -D, --tx-delay=    Tx delay (in usec.) each 10 loops (default 100us)\n"
        "  -E, --extended     Enables the extended ID support\n"
        "  -d, --device=      Interface to use (e.g. can0)\n"
        "  -l, --loops=       Number of test loops to execute\n"
        "  -T, --timeout=     Max read timeout\n"
        "  -s, --secure=      Secure mode, tx check rx queue\n"
        "  -v, --verbosity=   Verbosity level (3: loud | 0: quiet)\n"
        "  -V, --version      Show version and exit\n"
        "  -h, --help         Display usage information\n\n",
        APP_NAME, APP_NAME, APP_VERSION);
}


static struct opts_t *process_options(int argc, char *argv[])
{
    int opt_index, opt;
    static const char *short_options = "mD:d:Ed:T:l:v:Vsh";
    struct opts_t *retval;

    static const struct option long_options[] = {
        {"master", no_argument, NULL, 'm'},
        {"tx-delay", required_argument, NULL, 'D'},
        {"extended", no_argument, NULL, 'E'},
        {"device", required_argument, NULL, 'd'},
        {"timeout", required_argument, NULL, 'T'},
        {"loops", required_argument, NULL, 'l'},
        {"verbosity", required_argument, NULL, 'v'},
        {"secure", no_argument, NULL, 's'},
        {"version", no_argument, NULL, 'V'},
        {"help", no_argument, NULL, 'h'},
        {0, 0, 0, 0},
    };
    selected_verbosity = 2;

    /* Allocate the space for the internal data structure */
    retval = calloc(1, sizeof(struct opts_t));
    if (!retval) {
        pr_error("calloc failed, %s\n", strerror(errno));
        return NULL;
    }

    // Initialize CAN struct with some default values
    retval->timeout=5;
    retval->xdelay=100; // default delay 100us

    for (opt_index = 0;;) {
        opt = getopt_long(argc, argv, short_options, long_options, &opt_index);
        if (opt == EOF)
            break;

        switch (opt) {
        case 'm':
            retval->mode = MODE_TRANSMITTER;
            break;

        case 'D':
            retval->xdelay = atoi(optarg);
            break;

        case 'd':
            retval->iface = optarg;
            break;

        case 'v':
            selected_verbosity = atoi(optarg);
            break;

        case 'l':
            retval->loops = atol(optarg);
            break;

        case 'T':
            retval->timeout = atol(optarg);
            break;

        case 's':
            retval->secure = 1;
            break;

        case 'E':
            retval->extended = 1;
            break;

        case 'V':
            print_version();
            goto err_free_mem;

        case '?':
            fprintf(stderr, "Unknown option -- %c\n", opt);

        case 'h':
        default:
            print_usage();
            goto err_free_mem;
        }
    }

    if (!retval->loops) {
        pr_error("A valid number of test loops is required\n");
        goto err_free_mem;
    }

    if (!retval->iface) {
        pr_error("Need a CAN device for the test\n");
        goto err_free_mem;
    }

    /* Print the information about the started mode */
    if (retval->mode == MODE_TRANSMITTER)
        pr_naked("Running the test as TRANSMITTER\n");
    else
        pr_naked("Running the test as RECEIVER\n");

    pr_info("Testing with %i loops\n", retval->loops);
    return retval;

 err_free_mem:
    free(retval->canids);
    free(retval->tx_rates);
    free(retval);

    return NULL;
}



void update_frame_values (unsigned char *pattern, int *bytes, canid_t *id) {
    *pattern = *pattern + 1;
    if (*pattern > 255) {
        *pattern = 0;
    }

    *bytes = *bytes+1;
    if (*bytes > 8) {
        *bytes = 1;
    }

    *id = *id+1;
    if (main_opts->extended) {
        if (*id == (1 << 29)) {
            *id = 0;
        }
    }
    else if (*id == (1 << 11)) {
        *id = 0;
    }
}



/*
 * Depending on the selected command line options, the below function configures
 * the CAN-frames in two different modes:
 * - With a modified incremented frame Id (RTR or extended ID)
 * - Modified data pattern (incremented with 0x11)
 */
static void update_frame(struct can_frame *frame, canid_t id,
             unsigned char pattern, int bytes)
{
    frame->can_id = (main_opts->extended) ? (id | CAN_EFF_FLAG) : (id);
    frame->can_dlc = bytes;
    memset(frame->data, pattern, bytes);
}

/* Dump the content of a CAN-frame */
inline static void dump_frame(struct can_frame *frame, const char *name)
{
    int cnt;

    pr_naked("%s: ID 0x%03x | DLC 0x%02x | DATA ", name, frame->can_id, frame->can_dlc);
    for (cnt = 0; cnt < frame->can_dlc; cnt++)
        pr_naked("0x%02x ", frame->data[cnt]);
    pr_naked("\n");
}

/* Return zero if the frames are equal, otherwise one */
static int compare_frames(struct can_frame *rcv, struct can_frame *exp)
{
    if (rcv->can_id != exp->can_id)
        return 1;

    if (rcv->can_dlc != exp->can_dlc)
        return 1;

    return memcmp(rcv->data, exp->data, rcv->can_dlc);
}

static int xmit_frame(int sock, struct can_frame *frame)
{
    return write(sock, frame, CAN_FRAME_SIZE);
}

static void *start_test() {
    int sock, fl;
    int family = PF_CAN, type = SOCK_RAW, proto = CAN_RAW;
    struct sockaddr_can addr;
    struct ifreq ifr;
    int loop, retop, bytes;
    canid_t id = 0;
    int loopback = 1; /* 0 = disabled, 1 = enabled (default) */
    int recv_own_msgs = 1; /* 0 = disabled (default), 1 = enabled */
    can_err_mask_t err_mask = 0;
    struct can_frame txframe, rxframe;
    struct can_filter rfilter;
    unsigned char pattern;
    int retryCounter = 0, sleepTime = 10000, increment = 20000;

    struct timeval timeout;
    int ret, timerCounter = 0;
    unsigned long int limitCountet;


    if (main_opts->secure) {
        loopback = 1; /* 0 = disabled, 1 = enabled (default) */
        recv_own_msgs = 1; /* 0 = disabled (default), 1 = enabled */
    }
    else {
        loopback = 0;
        recv_own_msgs = 0;
    }


    limitCountet = (main_opts->timeout*1000);
    pattern = 0;
    bytes = 1;
    if (main_opts->extended)
        // (1 << 29) = 536870912
        id = (canid_t) 536870911;
    else
        // (1 << 11) = 2048
        id = (canid_t) 2047;

    /* Create the socket first */
    if ((sock = socket(family, type, proto)) < 0) {
        pr_error("socket() failed, %s\n", strerror(errno));
        return NULL;
    }

    rfilter.can_id = id;
    rfilter.can_mask = 0x00000000U;
    retop = setsockopt(sock, SOL_CAN_RAW, CAN_RAW_FILTER, &rfilter, 8);
    if (retop) {
        pr_error("Socket setup failed, %s\n", strerror(errno));
        goto close_socket;
    }

    strcpy(ifr.ifr_name, main_opts->iface);
    retop = ioctl(sock, SIOCGIFINDEX, &ifr);
    if (retop < 0) {
        pr_error("The IOCTL for `%s' failed, %s\n", main_opts->iface, strerror(errno));
        goto close_socket;
    }

    addr.can_ifindex = ifr.ifr_ifindex;
    addr.can_family = family;

    setsockopt(sock, SOL_CAN_RAW, CAN_RAW_LOOPBACK, &loopback, sizeof(loopback));
    setsockopt(sock, SOL_CAN_RAW, CAN_RAW_RECV_OWN_MSGS, &recv_own_msgs, sizeof(recv_own_msgs));

//     err_mask = CAN_ERR_TX_TIMEOUT | CAN_ERR_BUSOFF | CAN_ERR_ACK | CAN_ERR_LOSTARB | CAN_ERR_CRTL | CAN_ERR_PROT | CAN_ERR_TRX |CAN_ERR_BUSERROR | CAN_ERR_RESTARTED;
//     err_mask = CAN_ERR_FLAG;
    err_mask = 0xFFFFFFFF;

    retop = setsockopt(sock, SOL_CAN_RAW, CAN_RAW_ERR_FILTER, &err_mask, sizeof(err_mask));
    if (retop) {
        pr_error("ERROR MASK setup failed, %s\n", strerror(errno));
        goto close_socket;
    }


    /* Configure the socket correctly */
    fl = fcntl(sock, F_GETFD);
    if (fcntl(sock, F_SETFD, fl | O_SYNC | O_NDELAY | O_NOCTTY | FD_CLOEXEC) == -1) {
        pr_error("Socket config failed, %s\n", strerror(errno));
        goto close_socket;
    }

    if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
        pr_error("bind() failed, %s\n", strerror(errno));
        goto close_socket;
    }


    for (loop = 1; loop <= main_opts->loops; loop++) {
        update_frame(&txframe, id, pattern, bytes);
        /* In the transmitter mode only send the CAN-frame */
        if (main_opts->mode == MODE_TRANSMITTER) {
            retop = xmit_frame(sock, &txframe);
            if (retop < 0) {
                if ((errno == ENOBUFS) && (timerCounter < limitCountet)) {
                    timerCounter += 1;
                    if (retryCounter == 0) {
                        pr_warning("Could not sent pkt %i (%s), retrying...\n", loop, strerror(errno));
                    }
                    retryCounter += 1;
                    sleepTime += increment;
                    // Decrease the counter because we could not send the packet in this loop
                    loop--;
                    // Wait a moment until buffer is free
                    // @TODO, check if there is a function that allow us to check if there is space on buffer
                    usleep(sleepTime);
                    continue;
                }
                else {
                    if (errno != ENOBUFS)
                        pr_error("write function has returned error, %s\n", strerror(errno));
                    else
                        pr_error("Aborting by timeout, last pk transmited is %i\n", (loop - 1));
                    goto close_socket;
                }

            } else if (retop != CAN_FRAME_SIZE) {
                pr_error("Couldn't send the %i bytes (%i sent)\n",
                     CAN_FRAME_SIZE, retop);
                goto close_socket;
            } else {
                if (retryCounter != 0) {
                    pr_info("Packet %i sent after %i retries\n", loop, retryCounter);
                }
                retryCounter = 0;
                pr_debug("Pkt %i sent: ID 0x%03x Pattern 0x%02x DLC %i\n", loop, id, pattern, bytes);
            }

            if (main_opts->secure) {
                // read timeout
                fd_set readset;
                FD_ZERO(&readset);
                FD_SET(sock, &readset);
                if (main_opts->timeout != 0) {
                    timeout.tv_sec = 1;
                    timeout.tv_usec = 0;
                }
                ret = select(sock+1, &readset, NULL, NULL, &timeout);
                if (ret < 0) {
                    pr_error("select failed, %s\n", strerror(errno));
                    goto close_socket;
                } else if (ret == 0) {
                    pr_error("Read timeout after %i seconds, last pk received was %i\n", main_opts->timeout, (loop - 1));
                    goto close_socket;
                }
                retop = read(sock, &rxframe, CAN_FRAME_SIZE);
                if (retop < 0) {
                    pr_error("Read of CAN frame failed (loop %i), %s\n",
                             loop, strerror(errno));
                    goto close_socket;
                }
                if (rxframe.can_id & CAN_ERR_FLAG) {
                    pr_error("CAN ERROR FRAME READED!!!!");
                    goto close_socket;
                }
                if (compare_frames(&txframe, &rxframe)) {
                    pr_error("[LOOPBACK] Different CAN-frames at loop %i\n", loop);
                    dump_frame(&txframe, "\t* EXPECTED");
                    dump_frame(&rxframe, "\t* RECEIVED");
                    goto close_socket;
                }
                else
                    pr_debug("[LOOPBACK] frame %i with ID 0x%x received\n", loop, rxframe.can_id);

            }

            if (sleepTime > 10000)
                sleepTime -= increment;
            timerCounter = 0;

            // Apply delay every 10 loops
            if (loop%10 == 0) {
                if (main_opts->xdelay)
                    usleep(main_opts->xdelay);
            }


        } else {
            // read timeout
            fd_set readset;
            FD_ZERO(&readset);
            FD_SET(sock, &readset);

            if (main_opts->timeout != 0) {
                timeout.tv_sec = main_opts->timeout;
                timeout.tv_usec = 0;
            }

            ret = select(sock+1, &readset, NULL, NULL, &timeout);
            if (ret < 0) {
                pr_error("select failed, %s\n", strerror(errno));
                goto close_socket;
            } else if (ret == 0) {
                pr_error("Read timeout after %i seconds, last pk received was %i\n", main_opts->timeout, (loop - 1));
                goto close_socket;
            }

            retop = read(sock, &rxframe, CAN_FRAME_SIZE);
            if (retop < 0) {
                pr_error("Read of CAN frame failed (loop %i), %s\n",
                     loop, strerror(errno));
                goto close_socket;
            }

            if (rxframe.can_id & CAN_ERR_FLAG) {
                pr_error("CAN ERROR FRAME READED!!!!");
                goto close_socket;
            }

            pr_debug("Frame %i with ID 0x%x received\n", loop, rxframe.can_id);
            if (compare_frames(&txframe, &rxframe)) {
                pr_error("Different CAN-frames at loop %i\n", loop);
                dump_frame(&txframe, "\t* EXPECTED");
                dump_frame(&rxframe, "\t* RECEIVED");
                goto close_socket;
            }
        }
        update_frame_values(&pattern, &bytes, &id);
    }
    pr_info("Test completed\n");

 close_socket:
    close(sock);
    return 0;
}


int main(int argc, char **argv) {
//     int retval;
    /* Create the internal options */
    if (!(main_opts = process_options(argc, argv)))
        return EXIT_FAILURE;

    start_test();

    free(main_opts->tx_rates);
    free(main_opts->canids);
    free(main_opts);

    return EXIT_SUCCESS;
}
